/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include "cstrvec.h"

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(push)
#  pragma warning(disable : 5045) // Information that MS compiler will insert Spectre mitigation if /Qspectre switch specified
#endif

//////////////////////////////////////////////////////////
// vvv  based on char  vvv

cstr cs_join(ccvec strvec, const char *delims)
{
  const size_t cnt = cv_size(strvec);
  cstr ret = cs_init_zero(cnt << 3);
  ccstr *svec = PTRTOCCSTR(strvec);
  cs_assign_n(&ret, *svec, cs_size(*svec));
  if (cnt > 1U)
  {
    const size_t delims_len = strlen(delims);
    for (ccstr *it = svec + 1, *end = svec + cnt; it < end; ++it)
    {
      cs_append_n(&ret, delims, delims_len);
      cs_append_n(&ret, *it, cs_size(*it));
    }
  }

  return ret;
}

cvec cs_split(ccstr cs, const char *delims, size_t maxtok)
{
  cvec ret = cv_init_zero(0U, sizeof(cstr));
  if (maxtok)
  {
    const size_t delims_len = strlen(delims);
    const char *begin = cs, *found = NULL;
    for (size_t n = 1U; n < maxtok && (found = strstr(begin, delims)) != NULL; ++n)
    {
      cv_push_back(&ret, MAKECSTRPTR(cs_init_n(begin, (size_t)(found - begin))));
      begin = found + delims_len;
    }

    cv_push_back(&ret, MAKECSTRPTR(cs_init(begin)));
  }

  return ret;
}

cvec cstrvec_rdlines(FILE *stream, size_t maxlines)
{
  cvec ret = cv_init_zero(0U, sizeof(cstr));
  for (size_t n = 0U; n < maxlines && !feof(stream) && !ferror(stream); cv_push_back(&ret, MAKECSTRPTR(cs_init_rdline(stream))), ++n)
    ;

  return ret;
}

void cstrvec_free(ccvec strvec)
{
  for (ccstr *it = PTRTOCCSTR(strvec), *end = it + cv_size(strvec); it < end; cs_free(*it++))
    ;

  cv_free(strvec);
}

//////////////////////////////////////////////////////////
// vvv  based on wchar_t  vvv

cstrw csw_join(ccvec strwvec, const wchar_t *delimsw)
{
  const size_t cnt = cv_size(strwvec);
  cstrw ret = csw_init_zero(cnt << 3);
  ccstrw *svec = PTRTOCCSTRW(strwvec);
  csw_assign_n(&ret, *svec, csw_size(*svec));
  if (cnt > 1U)
  {
    const size_t delims_len = wcslen(delimsw);
    for (ccstrw *it = svec + 1, *end = svec + cnt; it < end; ++it)
    {
      csw_append_n(&ret, delimsw, delims_len);
      csw_append_n(&ret, *it, csw_size(*it));
    }
  }

  return ret;
}

cvec csw_split(ccstrw csw, const wchar_t *delimsw, size_t maxtok)
{
  cvec ret = cv_init_zero(0U, sizeof(cstrw));
  if (maxtok)
  {
    const size_t delims_len = wcslen(delimsw);
    const wchar_t *begin = csw, *found = NULL;
    for (size_t n = 1U; n < maxtok && (found = wcsstr(begin, delimsw)) != NULL; ++n)
    {
      cv_push_back(&ret, MAKECSTRWPTR(csw_init_n(begin, (size_t)(found - begin))));
      begin = found + delims_len;
    }

    cv_push_back(&ret, MAKECSTRWPTR(csw_init(begin)));
  }

  return ret;
}

cvec cstrwvec_rdlines(FILE *stream, size_t maxlines)
{
  cvec ret = cv_init_zero(0U, sizeof(cstrw));
  for (size_t n = 0U; n < maxlines && !feof(stream) && !ferror(stream); cv_push_back(&ret, MAKECSTRWPTR(csw_init_rdline(stream))), ++n)
    ;

  return ret;
}

void cstrwvec_free(ccvec strwvec)
{
  for (ccstrw *it = PTRTOCCSTRW(strwvec), *end = it + cv_size(strwvec); it < end; csw_free(*it++))
    ;

  cv_free(strwvec);
}

//////////////////////////////////////////////////////////
// vvv  strings and vectors of bytes  vvv

cstr cs_from_cvbyte(ccvec cv)
{
  return cs_init_n((const char *)cv, cv_size(cv));
}

cvec cvbyte_from_cs(ccstr cs)
{
  return cv_init((const void *)cs, cs_size(cs), (size_t)1U);
}

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(pop) // C5045 Spectre mitigation
#endif
