/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#ifndef CSTRVEC_H_INCLUDED__
#define CSTRVEC_H_INCLUDED__

#include "cstr.h"
#include "cvec.h"

//////////////////////////////////////////////////////////
// vvv  based on char  vvv

/** \brief  Macro to cast a pointer to void (cvec or ccvec) to a pointer to cstr. */
#define PTRTOCSTR(_cvec) PTRTO(_cvec, cstr)

/** \brief  Macro to cast a pointer to void (cvec or ccvec) to a pointer to ccstr. */
#define PTRTOCCSTR(_cvec) PTRTO(_cvec, ccstr)

/** \brief  Macro to create a pointer to cstr, decayed from an array literal out of the cstr values passed to the ellipsis. */
#define MAKECSTRPTR(...) MAKEPTR(cstr, __VA_ARGS__)

/** \brief  Concatenate the elements of a cvec of cstr in a new cstr.
 * \param  strvec  The cvec of cstr tokens.
 * \param  delims  Delimiter string to be inserted between the tokens.
 * \return  The cstr containing the joined tokens. */
cstr cs_join(ccvec strvec, const char *delims);

/** \brief  Tokenize a cstr into a cvec of cstr.
 * \param  cs      The cstr.
 * \param  delims  Delimiter string that divide the tokens in cs.
 * \param  maxtok  Maximum number of tokens to be created.
 *                 Pass -1 to process all tokens.
 * \return  The cvec containing the tokens as array of cstr. */
cvec cs_split(ccstr cs, const char *delims, size_t maxtok);

/** \brief  Create and initialize a cvec of cstr from a lines of characters read out of a file.
 * \param  stream    File stream to be read, a line ends at an occurrence of line feed or EOF.
 *                   Can be `stdin` if maxlines is not -1 when redirected from keyboard.
 * \param  maxlines  Maximum number of lines to be read.
 *                   Pass -1 to process all lines.
 * \return  Newly allocated cvec of cstr containing the copied lines without the line feed
 *          at the ent of the lines. */
cvec cstrvec_rdlines(FILE *stream, size_t maxlines);

/** \brief  Convenience function to delete a cvec of cstr including its cstr elements.
 * \param  strvec  The cvec of cstr to be destroyed. */
void cstrvec_free(ccvec strvec);

//////////////////////////////////////////////////////////
// vvv  based on wchar_t  vvv

/** \brief  Macro to cast a pointer to void (cvec or ccvec) to a pointer to cstrw. */
#define PTRTOCSTRW(_cvec) PTRTO(_cvec, cstrw)

/** \brief  Macro to cast a pointer to void (cvec or ccvec) to a pointer to ccstrw. */
#define PTRTOCCSTRW(_cvec) PTRTO(_cvec, ccstrw)

/** \brief  Macro to create a pointer to cstrw, decayed from an array literal out of the cstrw values passed to the ellipsis. */
#define MAKECSTRWPTR(...) MAKEPTR(cstrw, __VA_ARGS__)

/** \brief  Concatenate the elements of a cvec of cstrw in a new cstrw.
 * \param  strwvec  The cvec of cstrw tokens.
 * \param  delimsw  Delimiter string to be inserted between the tokens.
 * \return  The cstrw containing the joined tokens. */
cstrw csw_join(ccvec strwvec, const wchar_t *delimsw);

/** \brief  Tokenize a cstrw into a cvec of cstrw.
 * \param  csw      The cstrw.
 * \param  delimsw  Delimiter string that divide the tokens in csw.
 * \param  maxtok   Maximum number of tokens to be created.
 *                  Pass -1 to process all tokens.
 * \return  The cvec containing the tokens as array of cstrw. */
cvec csw_split(ccstrw csw, const wchar_t *delimsw, size_t maxtok);

/** \brief  Create and initialize a cvec of cstrw from a lines of characters read out of a file.
 * \param  stream    File stream to be read, a line ends at an occurrence of line feed or EOF.
 *                   Can be `stdin` if maxlines is not -1 when redirected from keyboard.
 * \param  maxlines  Maximum number of lines to be read.
 *                   Pass -1 to process all lines.
 * \return  Newly allocated cvec of cstr containing the copied lines without the line feed
 *          at the ent of the lines. */
cvec cstrwvec_rdlines(FILE *stream, size_t maxlines);

/** \brief  Convenience function to delete a cvec of cstrw including its cstrw elements.
 * \param  strwvec  The cvec of cstrw to be destroyed. */
void cstrwvec_free(ccvec strwvec);

//////////////////////////////////////////////////////////
// vvv  strings and vectors of bytes  vvv

/** \brief  Create and initialize a cstr from cvec of bytes.
 * \param  cv  cvec of bytes.
 * \return  Newly allocated cstr containing the copied data from cv and a terminating null. */
cstr cs_from_cvbyte(ccvec cv);

/** \brief  Create and initialize a cvec of bytes from cstr.
 * \param  cs  The cstr.
 * \return  Newly allocated cvec of bytes containing the copied data from cs without
 *          the terminating null. */
cvec cvbyte_from_cs(ccstr cs);

#endif // CSTRVEC_H_INCLUDED__
