/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include "cvec.h"

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(push)
#  pragma warning(disable : 5045) // Information that MS compiler will insert Spectre mitigation if /Qspectre switch specified
#endif

// A rather abstract structure. It's never instantiated but makes it easier to access the information in the allocated memory space.
struct cvec_base
{
  size_t capacity; // number of bytes allocated
  size_t size; // number of elements
  size_t elementsize; // size of an element in bytes
  unsigned char data[sizeof(size_t)]; // Placeholder for the actual cvec.
};

// Actually only these pointers to cvec_base are used.
typedef struct cvec_base *pcvbase;
typedef const struct cvec_base *pccvbase;

// Distance in bytes of the data member from the beginning of cvec_base.
#define CVHEADSIZE (offsetof(struct cvec_base, data))

// Requested capacity is clamped to the double of this value.
#define CVHALFSIZE_MAX (((size_t)PTRDIFF_MAX - CVHEADSIZE) >> 1)

// Convenience functions which calculate the address of the beginning of a cvec_base from its data member, and casts it to the required pointer type.
static inline pcvbase get_cvbase_addr(cvec cv)
{
  return (pcvbase)(void *)((unsigned char *)cv - CVHEADSIZE); // The intermediate cast to void* suppresses the warning that the alignment increases from 1 to sizeof(size_t).
}

static inline pccvbase get_ccvbase_addr(ccvec cv)
{
  return (pccvbase)(const void *)((const unsigned char *)cv - CVHEADSIZE);
}

// Execution of CVDIE causes the program to exit abnormally.
#if defined(_MSC_VER)
#  define CVDIE __debugbreak()
#elif defined(__clang__) || defined(__GNUC__)
#  define CVDIE __builtin_trap()
#else
#  define CVDIE ((void)(*(volatile char *)0 = 0))
#endif

// (Re-)allocate the necessary memory, die if the allocation failed.
static pcvbase cv_alloc_or_die(pcvbase ptr, size_t req)
{
  if (ptr != NULL && req == (CVHALFSIZE_MAX << 1) && ptr->capacity == req)
    CVDIE; // The maximum possible memory has been already allocated. Another call of this function won't return more memory anyway.

  const pcvbase ret = (pcvbase)realloc(ptr, CVHEADSIZE + req);
  if (!ret)
    CVDIE; // Not so nice. But not getting memory anymore means that your OS is close to collapse. Better immediately give it a chance to recover.

  ret->capacity = req;
  return ret;
}

cvec cv_init(const void *seq, size_t cnt, size_t elementsize)
{
  const size_t req = cnt > 31U ? (cnt * elementsize > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : cnt * elementsize) << 1 : 64U * elementsize;
  const pcvbase ptr = cv_alloc_or_die(NULL, req);
  ptr->size = cnt;
  ptr->elementsize = elementsize;
  memcpy(ptr->data + 0, seq, cnt * elementsize); // +0 to decay the array to a pointer; analysis tools may consider the possibility to violate the declared array boundary (which is intentionally the case)
  return ptr->data;
}

cvec cv_init_zero(size_t size, size_t elementsize)
{
  const size_t req = size > 31U ? (size * elementsize > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : size * elementsize) << 1 : 64U * elementsize;
  const pcvbase ptr = cv_alloc_or_die(NULL, req);
  ptr->size = 0U;
  ptr->elementsize = elementsize;
  return ptr->data;
}

void cv_free(ccvec cv)
{
  // Work around the ancient quirk in the C lib where the parameter of free() is `void*` rather than `const void*`.
  free(discard_const_impl_detail(get_ccvbase_addr(cv)));
}

cvec cv_append(cvec *pcv, const void *seq, size_t cnt)
{
  if (!cnt)
    return *pcv;

  pcvbase ptr = get_cvbase_addr(*pcv);
  const size_t req = (ptr->size + cnt) * ptr->elementsize;
  if (req > ptr->capacity)
    ptr = cv_alloc_or_die(ptr, (req > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : req) << 1);

  memcpy(ptr->data + (ptr->size * ptr->elementsize), seq, cnt * ptr->elementsize);
  ptr->size += cnt;
  *pcv = ptr->data;
  return ptr->data;
}

cvec cv_assign(cvec *pcv, const void *seq, size_t cnt)
{
  pcvbase ptr = get_cvbase_addr(*pcv);
  const size_t req = cnt * ptr->elementsize;
  if (req > ptr->capacity)
    ptr = cv_alloc_or_die(ptr, (req > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : req) << 1);

  memcpy(ptr->data + 0, seq, req);
  ptr->size = cnt;
  *pcv = ptr->data;
  return ptr->data;
}

const void *cv_at(ccvec cv, size_t pos)
{
  const pccvbase ptr = get_ccvbase_addr(cv);
  if (pos >= ptr->size)
    return NULL;

  return ptr->data + (pos * ptr->elementsize);
}

size_t cv_capacity(ccvec cv)
{
  const pccvbase ptr = get_ccvbase_addr(cv);
  return ptr->capacity / ptr->elementsize;
}

cvec cv_clear(cvec cv)
{
  const pcvbase cur = get_cvbase_addr(cv);
  cur->size = 0U;
  return cv;
}

bool cv_empty(ccvec cv)
{
  const pccvbase ptr = get_ccvbase_addr(cv);
  return !ptr->size;
}

cvec cv_erase(cvec cv, size_t pos, size_t cnt)
{
  const pcvbase ptr = get_cvbase_addr(cv);
  if (pos >= ptr->size)
    return cv;

  const size_t diff = ptr->size - pos,
               len = (diff < cnt ? diff : cnt);

  unsigned char *const cursor = (unsigned char *)cv + pos;
  memmove(cursor, cursor + (len * ptr->elementsize), (diff - len) * ptr->elementsize);
  ptr->size -= len;
  return cv;
}

bool cv_for_each(cvec cv, size_t pos, size_t cnt, cv_for_each_proc_t callback_func, void *user_param)
{
  const pcvbase ptr = get_cvbase_addr(cv);
  if (!cnt || pos >= ptr->size)
    return false;

  for (unsigned char *it = ptr->data + pos * ptr->elementsize, *const end = (cnt == SIZE_MAX || pos + cnt >= ptr->size) ? ptr->data + (ptr->size * ptr->elementsize) : it + (cnt * ptr->elementsize);
       it < end && callback_func(it, user_param);
       it += ptr->elementsize)
    ;

  return true;
}

cvec cv_insert(cvec *pcv, size_t pos, const void *seq, size_t cnt)
{
  pcvbase ptr = get_cvbase_addr(*pcv);
  if (pos >= ptr->size)
    return cv_append(pcv, seq, cnt);

  const size_t req = (ptr->size + cnt) * ptr->elementsize;
  if (req > ptr->capacity)
    ptr = cv_alloc_or_die(ptr, (req > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : req) << 1);

  unsigned char *const cursor = ptr->data + (pos * ptr->elementsize);
  memmove(cursor + (cnt * ptr->elementsize), cursor, (ptr->size - pos) * ptr->elementsize);
  memcpy(cursor, seq, cnt * ptr->elementsize);
  ptr->size += cnt;
  *pcv = ptr->data;
  return ptr->data;
}

const void *cv_pop_back(cvec cv)
{
  const pcvbase ptr = get_cvbase_addr(cv);
  if (!ptr->size)
    return NULL;

  return ptr->data + (--ptr->size * ptr->elementsize);
}

cvec cv_push_back(cvec *pcv, const void *elem)
{
  return cv_append(pcv, elem, 1U);
}

cvec cv_reserve(cvec *pcv, size_t capa)
{
  pcvbase ptr = get_cvbase_addr(*pcv);
  if (ptr->capacity / ptr->elementsize >= capa)
    return *pcv;

  const size_t req = capa * ptr->elementsize;
  ptr = cv_alloc_or_die(ptr, (req > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : req) << 1);
  *pcv = ptr->data;
  return ptr->data;
}

cvec cv_resize(cvec *pcv, size_t size, const void *elem)
{
  pcvbase ptr = get_cvbase_addr(*pcv);
  if (ptr->size == size)
    return *pcv;

  if (ptr->size > size)
  {
    ptr->size = size;
    return *pcv;
  }

  const size_t req = size * ptr->elementsize;
  if (req > ptr->capacity)
    ptr = cv_alloc_or_die(ptr, (req > CVHALFSIZE_MAX ? CVHALFSIZE_MAX : req) << 1);

  if (!elem)
  {
    unsigned char *cursor = ptr->data + (ptr->size * ptr->elementsize);
    memset(cursor, 0, (size_t)(ptr->data + (size * ptr->elementsize) - cursor));
  }
  else
  {
    for (unsigned char *it = ptr->data + (ptr->size * ptr->elementsize), *const end = ptr->data + req; it < end; it += ptr->elementsize)
      memcpy(it, elem, ptr->elementsize);
  }

  ptr->size = size;
  *pcv = ptr->data;
  return ptr->data;
}

cvec cv_shrink_to_fit(cvec *pcv)
{
  pcvbase ptr = get_cvbase_addr(*pcv);
  const size_t req = ptr->size * ptr->elementsize;
  if (ptr->capacity > req)
    ptr = cv_alloc_or_die(ptr, req);

  *pcv = ptr->data;
  return ptr->data;
}

cvec cv_update_size(cvec cv, size_t size)
{
  const pcvbase ptr = get_cvbase_addr(cv);
  const size_t capa = ptr->capacity / ptr->elementsize;
  ptr->size = size > capa ? capa : size;
  return cv;
}

size_t cv_size(ccvec cv)
{
  const pccvbase ptr = get_ccvbase_addr(cv);
  return ptr->size;
}

#if defined(_MSC_VER) && _MSC_VER >= 1914
#  pragma warning(pop) // C5045 Spectre mitigation
#endif
