/*
Copyright (c) 2020 Steffen Illhardt

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

// NOTE: The C standard reserved function names beginning with mem, str, and wcs
// for future use. However, the C string lib has not been updated for decades.
// If you are worried about using these functions then just don't. Also feel
// free to change the function names. It's one of the advantages of open source.

#ifndef MEMMEM_H_INCLUDED__
#define MEMMEM_H_INCLUDED__

#include <stddef.h>
#include <string.h>
#include <wchar.h>

/** \brief  Find the first occurrence of the target byte sequence in the block of memory pointed to by buffer
 *          using the Rabin-Karp algorithm. This function is similar to the `strstr` standard function,
 *          but is suitable for binary data, too.
 * \param  buffer       Pointer to the block of memory where the search is performed.
 * \param  buffer_size  Number of bytes in buffer to be analyzed.
 * \param  target       Pointer to the block of memory to search for.
 * \param  target_size  Number of bytes in target to be searched.
 * \return  Pointer to the first byte in buffer where the target sequence is found,
 *          or NULL if the target sequence is not found. */
void *memmem(const void *buffer, size_t buffer_size, const void *target, size_t target_size);

/** \brief  Find the last occurrence of the target byte sequence in the block of memory pointed to by buffer
 *          using the Rabin-Karp algorithm.
 * \param  buffer       Pointer to the block of memory where the search is performed.
 * \param  buffer_size  Number of bytes in buffer to be analyzed.
 * \param  target       Pointer to the block of memory to search for.
 * \param  target_size  Number of bytes in target to be searched.
 * \return  Pointer to the first byte in buffer where the target sequence is found,
 *          or NULL if the target sequence is not found. */
void *memrmem(const void *buffer, size_t buffer_size, const void *target, size_t target_size);

/** \brief  Find the last occurrence of the substr in the string pointed to by str.
 * \param  str     Nullterminated string of characters where the search is performed.
 * \param  substr  Nullterminated string of characters to search for.
 * \return  Pointer to the first character in str where the substr is found,
 *          or NULL if the substr is not found. */
char *strrstr(const char *str, const char *substr);

/** \brief  Find the last occurrence of the substrw in the string pointed to by strw.
 * \param  strw     Nullterminated string of wide characters where the search is performed.
 * \param  substrw  Nullterminated string of wide characters to search for.
 * \return  Pointer to the first wide character in strw where the substrw is found,
 *          or NULL if the substrw is not found. */
wchar_t *wcsrstr(const wchar_t *strw, const wchar_t *substrw);

#endif // MEMMEM_H_INCLUDED__
